# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/9/6


from dupfilter.filters import decorate_warning
from dupfilter.filters.redis import RedisFilter

EXISTS_SCRIPT = """
local keys = KEYS
local values = ARGV
local result = {}
for index, key in pairs(keys) do
    local exist = redis.call('SISMEMBER', key, values[index])
    table.insert(result, exist)
end
return result
"""
INSERT_SCRIPT = """
local keys = KEYS
local values = ARGV
for index, key in pairs(keys) do
    redis.call('SADD', key, values[index])
end
return true
"""
EXISTS_AND_INSERT_SCRIPT = """
local keys = KEYS
local values = ARGV
local result = {}
for index, key in pairs(keys) do
    local exist = redis.call('SADD', key, values[index])
    if exist==0 then
        table.insert(result, 1)
    else
        table.insert(result, 0)
    end
end
return result
"""
RESET_SCRIPT = """
local key = KEYS[1]
local count = ARGV[1]
redis.call('SPOP', key, count)
return true
"""


class RedisSetFilter(RedisFilter):
    def __init__(self, server, key, block_num=1, *args, **kwargs):
        self.key = key
        self.block_num = block_num
        super(RedisSetFilter, self).__init__(server, *args, **kwargs)
        self._exists_script = self.server.register_script(EXISTS_SCRIPT)
        self._insert_script = self.server.register_script(INSERT_SCRIPT)
        self._exists_and_insert_script = self.server.register_script(
            EXISTS_AND_INSERT_SCRIPT)

    @decorate_warning
    def exists(self, value):
        return self.exists_many([value])[0]

    @decorate_warning
    def exists_many(self, values):
        keys, new_values = self._get_keys_and_values(values)
        stats = self._exists_script(keys=keys, args=new_values)
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, new_values, stats)
        return stats

    @decorate_warning
    def exists_and_insert(self, value):
        return self.exists_and_insert_many([value])[0]

    @decorate_warning
    def exists_and_insert_many(self, values):
        keys, new_values = self._get_keys_and_values(values)
        stats = self._exists_and_insert_script(keys=keys, args=new_values)
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, new_values, stats)
        return stats

    @decorate_warning
    def insert(self, value):
        return self.insert_many([value])[0]

    @decorate_warning
    def insert_many(self, values):
        keys, values = self._get_keys_and_values(values)
        stat = self._insert_script(keys=keys, args=values)
        return [bool(stat) for _ in range(len(values))]

    def _get_keys_and_values(self, values):
        values = [self._value_hash(value) for value in values]
        keys = [self.key + self._get_block(value) for value in values]
        values = [self._value_compress(value) for value in values]
        return keys, values

    def _get_block(self, value):
        if self.block_num == 1:
            return ''
        return str(int(value, 16) % self.block_num)


class AsyncRedisSetFilter(RedisSetFilter):
    @decorate_warning
    async def insert(self, value):
        stats = await self.insert_many([value])
        return stats[0]

    @decorate_warning
    async def insert_many(self, values):
        keys, values = self._get_keys_and_values(values)
        stat = await self._insert_script(keys=keys, args=values)
        return [bool(stat) for _ in range(len(values))]

    @decorate_warning
    async def exists(self, value):
        stats = await self.exists_many([value])
        return stats[0]

    @decorate_warning
    async def exists_many(self, values):
        keys, new_values = self._get_keys_and_values(values)
        stats = await self._exists_script(keys=keys, args=new_values)
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, new_values, stats)
        return stats

    @decorate_warning
    async def exists_and_insert(self, value):
        stats = await self.exists_and_insert([value])
        return stats[0]

    @decorate_warning
    async def exists_and_insert_many(self, values):
        keys, new_values = self._get_keys_and_values(values)
        stats = await self._exists_and_insert_script(keys=keys, args=values)
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, new_values, stats)
        return stats


# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/9/6


from dupfilter import utils
from dupfilter.filters import decorate_warning
from dupfilter.filters.redis import RedisFilter

EXISTS_SCRIPT = """
local keys = KEYS
local offsets = ARGV
local result = {}
for index, key in pairs(keys) do
    local exist = 1
    for offset in string.gmatch(offsets[index], "[^,]+") do
        exist = redis.call('GETBIT', key, offset)
        if exist == 0 then
            break
        end
    end
    table.insert(result, exist)
end
return result
"""

INSERT_SCRIPT = """
local keys = KEYS
local offsets = ARGV
for index, key in pairs(keys) do
    for offset in string.gmatch(offsets[index], "[^,]+") do
        redis.call('SETBIT', key, offset, 1)
    end
end
return true
"""

EXISTS_AND_INSERT_SCRIPT = """
local keys = KEYS
local offsets = ARGV
local result = {}
for index, key in pairs(keys) do
    local exist = 1
    for offset in string.gmatch(offsets[index], "[^,]+") do
        local _exist = redis.call('GETBIT', key, offset)
        if _exist == 0 then
            exist = _exist
            redis.call('SETBIT', key, offset, 1)
        end
    end
    table.insert(result, exist)
end
return result
"""

RESET_SCRIPT = """
local key = KEYS[1]
local offsets = ARGV
for _, offset in pairs(offsets) do
    redis.call('SETBIT', key, offset, 0)
end
return true
"""


class SimpleHash(object):
    def __init__(self, bit, seed):
        self.bit = bit
        self.seed = seed

    def hash(self, value):
        ret = 0
        for i in range(len(value)):
            ret += self.seed * ret + ord(value[i])
        return self.bit & ret


class RedisBloomFilter(RedisFilter):
    def __init__(self, server, key, bit=32, hash_num=6, block_num=1,
                 *args, **kwargs):
        """

        :param server: The redis obj, redis.Redis()
        :param key:
        :param bit:
        :param hash_num:
        :param block_num:
        :param reset: 是否进行重置/清除，
        :param reset_proportion: 达到比例后进行重置/清除，
        :param reset_check_period: 重置/清除比例获取计算周期，
        """
        self.key = key
        if bit > 32:
            raise ValueError('The bit value not > 32!')
        self.bit = (1 << bit) - 1  # Redis的String类型最大容量为512M，即bit=32
        self.hash_num = hash_num
        self.seeds = range(1, hash_num + 1)
        self.shs = [SimpleHash(self.bit, seed) for seed in self.seeds]
        self.block_num = block_num
        super(RedisBloomFilter, self).__init__(server, *args, **kwargs)
        self._exists_script = self.server.register_script(EXISTS_SCRIPT)
        self._insert_script = self.server.register_script(INSERT_SCRIPT)
        self._exists_insert_script = self.server.register_script(
            EXISTS_AND_INSERT_SCRIPT)

    @decorate_warning
    def exists(self, value):
        return self.exists_many([value])[0]

    @decorate_warning
    def exists_many(self, values):
        keys, offsets, new_values = self._get_keys_and_offsets(values)
        stats = self._exists_script(keys=keys, args=offsets)
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, new_values, stats)
        return stats

    @decorate_warning
    def insert(self, value):
        return self.insert_many([value])[0]

    @decorate_warning
    def insert_many(self, values):
        keys, offsets, new_values = self._get_keys_and_offsets(values)
        stat = self._insert_script(keys=keys, args=offsets)
        return [bool(stat) for _ in range(len(values))]

    @decorate_warning
    def exists_and_insert(self, value):
        return self.exists_and_insert_many([value])[0]

    @decorate_warning
    def exists_and_insert_many(self, values):
        keys, offsets, new_values = self._get_keys_and_offsets(values)
        stats = self._exists_insert_script(keys=keys, args=offsets)
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, new_values, stats)
        return stats

    def _get_keys_and_offsets(self, values):
        values = [self._value_hash(value) for value in values]
        keys = [self.key + self._get_block(value) for value in values]
        offsets = [','.join([str(sh.hash(
            self._value_compress(value))) for sh in self.shs]
        ) for value in values]
        return keys, offsets, values

    def _get_block(self, value):
        return str(int(value[0:2], 16) % self.block_num)


class AsyncRedisBloomFilter(RedisBloomFilter):
    @decorate_warning
    async def exists(self, value):
        stats = await self.exists_many([value])
        return stats[0]

    @decorate_warning
    async def exists_many(self, values):
        keys, offsets, new_values = self._get_keys_and_offsets(values)
        stats = await self._exists_script(keys=keys, args=offsets)
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, new_values, stats)
        return stats

    @decorate_warning
    async def insert(self, value):
        stats = await self.insert_many([value])
        return stats[0]

    @decorate_warning
    async def insert_many(self, values):
        keys, offsets, new_values = self._get_keys_and_offsets(values)
        stat = await self._insert_script(keys=keys, args=offsets)
        return [bool(stat) for _ in range(len(values))]

    @decorate_warning
    async def exists_and_insert(self, value):
        stats = await self.exists_and_insert_many([value])
        return stats[0]

    @decorate_warning
    async def exists_and_insert_many(self, values):
        keys, offsets, new_values = self._get_keys_and_offsets(values)
        stats = await self._exists_insert_script(keys=keys, args=offsets)
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, new_values, stats)
        return stats

if __name__ == '__main__':
    bf = BloomFilter({}, '1', value_compress_func=utils.hex2b64)
    bf.insert("123")



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/12/8


import time
from dupfilter.filters import Filter, decorate_warning


class SQLFilter(Filter):

    def __init__(self, connection, table,
                 record_time=False, *args, **kwargs):
        """
        需建立一张包含id、insert_time（可选，结合record_time逻辑）的表，id为主键
        :param connection:
        :param table:
        :param record_time: 是否记录插入时间
        :param args:
        :param kwargs:
        """
        self.table = table
        self.connection = connection
        self.cursor = self.connection.cursor()
        self.record_time = record_time
        super(SQLFilter, self).__init__(*args, **kwargs)
        self._create_table()

    def _create_table_sql(self):
        if not self.record_time:
            sql = """  
            CREATE TABLE IF NOT EXISTS %s (  
                id VARCHAR(32) NOT NULL,
                PRIMARY KEY (id)  
            )
            """ % self.table
        else:
            sql = """  
            CREATE TABLE IF NOT EXISTS %s (  
                id VARCHAR(32) NOT NULL,  
                insert_time INT ,
                PRIMARY KEY (id)  
            )
                """ % self.table
        return sql

    def _exists_table(self):
        return False

    def _create_table(self):
        if self._exists_table():
            return
        sql = self._create_table_sql()
        try:
            self.cursor.execute(sql)
            self.connection.commit()
            self.logger.info("创建去重表%s成功" % self.table)
        except Exception as e:
            self.logger.info("创建去重表%s失败" % self.table)

    @decorate_warning
    def exists(self, value):
        return self.exists_many([value])[0]

    def _exists_sql(self, values):
        values = [self._value_hash_and_compress(value) for value in values]
        s_values = str(tuple(values)) if len(
            values) > 1 else f"('{values[0]}')"
        sql = f"select id from {self.table} where id in {s_values}"
        return sql, values

    @decorate_warning
    def exists_many(self, values):
        sql, new_values = self._exists_sql(values)
        self.cursor.execute(sql)
        result = self.cursor.fetchall()
        try:
            result = [res[0] for res in result]
        except KeyError:
            result = [res['id'] for res in result]
        stats = [value in result for value in new_values]
        self._log_exists(values, new_values, stats)
        return stats

    @decorate_warning
    def insert(self, value):
        return self.insert_many([value])[0]

    def _insert_sql(self, values):
        values = [self._value_hash_and_compress(value) for value in values]
        if self.record_time:
            sql = f"""INSERT IGNORE INTO {self.table}
                     (id, insert_time) VALUES (%s, %s)"""
            values = [(value, int(time.time())) for value in values]
        else:
            sql = f"INSERT IGNORE INTO {self.table} (id) VALUES (%s)"
        return sql, values

    @decorate_warning
    def insert_many(self, values):
        sql, values = self._insert_sql(values)
        self.cursor.executemany(sql, values)
        self.connection.commit()
        return [True for _ in values]

    @decorate_warning
    def exists_and_insert(self, value):
        return self.exists_and_insert_many([value])[0]

    @decorate_warning
    def exists_and_insert_many(self, values):
        stats = self.exists_many(values)
        values = [value for stat, value in zip(stats, values) if not stat]
        self.insert_many(values)
        return stats

    def close(self):
        try:
            self.cursor.close()
            self.connection.close()
            self.logger.info("去重数据库连接关闭成功")
        except Exception as e:
            self.logger.warning("去重数据库连接关闭失败：%s" % str(e))


if __name__ == '__main__':
    import pymysql

    conn = pymysql.connect(host='192.168.1.10', user='root', password='123456',
                           database='test')
    cursor = conn.cursor()

    f = SQLFilter(conn, cursor, 'dup', record_time=False)
    r = f.exists_and_insert_many(['1', '2', '5'])
    # r = f.insert_many(['1', '2', '6'])
    print(r)



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/9/6

import hashlib
import base64

SFB_CHARACTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"


def md5(s):
    _md5 = hashlib.md5()
    _md5.update(s.encode('utf8'))
    return _md5.hexdigest()


def decimal2sfb(num):
    if num == 0:
        return ""
    result = decimal2sfb(num // 64)
    result += SFB_CHARACTERS[num % 64]
    return result


def hex2sfb(s):
    """
    自定义十六进制转六十四进制，速度比hex2b64慢10倍，但能压缩少1-3个字符
    :param s:
    :return:
    """
    return decimal2sfb(int(s, 16))


def hex2b64(s):
    """
    十六进制转base64
    :param s:
    :return:
    """
    return base64.b64encode(bytes.fromhex(s)).decode('utf-8')


if __name__ == '__main__':
    import time

    t1 = time.time()
    for i in range(10000):
        r = hex2sfb('698d51a19d8a121ce581499d7b701557')
    t2 = time.time()

    # 示例用法
    for i in range(10000):
        hex2b64('698d51a19d8a121ce581499d7b701557')
    t3 = time.time()
    print(t2 - t1, t3 - t2)



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/9/7
from dupfilter.filters import Filter


class RedisFilter(Filter):
    def __init__(self, server, *args, **kwargs):
        self.server = server
        super(RedisFilter, self).__init__(*args, **kwargs)



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/9/6

from dupfilter.filters.file import FileFilter
from dupfilter.filters.memory import MemoryFilter
from dupfilter.filters.redis.bloomfilter import RedisBloomFilter
from dupfilter.filters.redis.bloomfilter import AsyncRedisBloomFilter
from dupfilter.filters.redis.stringfilter import RedisStringFilter
from dupfilter.filters.redis.stringfilter import AsyncRedisStringFilter
from dupfilter.filters.redis.setfilter import RedisSetFilter
from dupfilter.filters.redis.setfilter import AsyncRedisSetFilter
from dupfilter.filters.redis.sortedsetfilter import RedisSortedSetFilter
from dupfilter.filters.redis.sortedsetfilter import AsyncRedisSortedSetFilter
from dupfilter.filters.sql import SQLFilter
from dupfilter.filters.sql.mysql import MySQLFilter
from dupfilter.filters.sql.mysql import AsyncMySQLFilter
from dupfilter.filters.sql.oracle import OracleSQLFilter
from dupfilter.filters.sql.oracle import AsyncOracleSQLFilter
from dupfilter.filters.sql.sqlite import SQLiteFilter
from dupfilter.filters import DefaultFilter
from dupfilter.filters import AsyncDefaultFilter
from dupfilter.filters import FilterCounter
from dupfilter.filters import PageFilterCounter



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/9/6


import asyncio
import functools
import logging

from cachetools import TTLCache

from dupfilter import utils


def decorate_warning(func):
    @functools.wraps(func)
    def wrapper(flt, value_or_values, *args, **kwargs):
        try:
            result = func(flt, value_or_values, *args, **kwargs)
        except Exception as e:
            if isinstance(value_or_values, list):
                result = [flt.default_stat for _ in value_or_values]
            else:
                result = flt.default_stat
            flt.logger.warning('去重操作失败，返回默认值：%s，错误详情：%s' % (
                flt.default_stat, str(e)))
        return result

    @functools.wraps(func)
    async def async_wrapper(flt, value_or_values, *args, **kwargs):
        try:
            result = await func(flt, value_or_values, *args, **kwargs)
        except Exception as e:
            if isinstance(value_or_values, list):
                result = [flt.default_stat for _ in value_or_values]
            else:
                result = flt.default_stat
            flt.logger.warning('去重操作失败，返回默认值：%s，错误详情：%s' % (
                flt.default_stat, str(e)))
        return result

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return wrapper


class Reset(object):
    def __init__(self, max_count, max_rate=0.8,
                 reset_to_rate=0.5, reset_type=None,
                 monitor_time=3600):
        self.max_count = max_count
        self.max_rate = max_rate
        self.reset_to_rate = reset_to_rate
        self.reset_type = reset_type
        self.monitor_time = monitor_time
        self.resetting = False
        self.timer = TTLCache(1, monitor_time)
        self.set_timer()
        self.flt = None

    def set_timer(self):
        self.timer['timer'] = 'timer'

    def get_timer(self):
        return self.timer.get('timer')

    @property
    def used_rate(self):
        return self.current_count / self.max_count

    @property
    def current_count(self):
        raise NotImplemented

    @property
    def reset_count(self):
        return self.current_count - int(self.max_count * self.reset_to_rate)

    def _reset(self):
        if self.resetting:
            return
        if self.get_timer():
            return
        if self.used_rate < self.max_rate:
            return
        self.resetting = True
        self.reset()
        if self.resetting:
            self.resetting = False
        self.set_timer()

    def reset(self):
        raise NotImplemented


def decorate_reset(func):
    @functools.wraps(func)
    def wrapper(flt, *args, **kwargs):
        try:
            if flt.reset:
                flt.reset._reset()
        except Exception as e:
            flt.reset.resetting = False
            flt.logger.warning("重置操作失败：%s" % str(e))
        result = func(flt, *args, **kwargs)
        return result

    @functools.wraps(func)
    async def async_wrapper(flt, *args, **kwargs):
        try:
            if flt.reset:
                await flt.reset._reset()
        except Exception as e:
            flt.reset.resetting = False
            flt.logger.warning("重置操作失败：%s" % str(e))
        result = await func(flt, *args, **kwargs)
        return result

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return wrapper


class Filter(object):
    def __init__(
            self,
            value_hash_func=utils.md5,
            value_compress_func=None,
            default_stat=False,
            logger=None,
            logger_level=logging.DEBUG,
            reset=None
    ):
        self.value_hash_func = value_hash_func or (lambda value: value)
        self.value_compress_func = value_compress_func or (lambda value: value)
        self.default_stat = bool(default_stat)
        if not logger:
            self.logger = logging.getLogger('dupfilter')
            self.logger.setLevel(logger_level)
            if not self.logger.handlers:
                ft = logging.Formatter(
                    fmt='%(asctime)s [%(name)s] %(levelname)s: %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S'
                )
                ch = logging.StreamHandler()
                ch.setFormatter(ft)
                ch.setLevel(logger_level)
                self.logger.addHandler(ch)
        else:
            self.logger = logger
        if reset:
            reset.flt = self
        self.reset = reset

    def _log_exists(self, initial_values, handle_values, stats):
        for mixed in zip(stats, handle_values, initial_values):
            self.logger.info('去重结果：%s，去重值：%s，原始值：%s' % (
                mixed[0], mixed[1], mixed[2]))

    def exists(self, value):
        raise NotImplemented

    def insert(self, value):
        raise NotImplemented

    def exists_many(self, values):
        raise NotImplemented

    def insert_many(self, values):
        raise NotImplemented

    def exists_and_insert(self, value):
        pass

    def exists_and_insert_many(self, values):
        pass

    def _value_hash(self, value):
        return self.value_hash_func(value)

    def _value_compress(self, value):
        return self.value_compress_func(value)

    def _value_hash_and_compress(self, value):
        return self.value_compress_func(self.value_hash_func(value))

    def close(self):
        pass


class DefaultFilter(object):
    def __init__(self, default_stat=False):
        self.default_stat = bool(default_stat)

    def exists(self, value):
        return self.default_stat

    def insert(self, value):
        return True

    def exists_many(self, values):
        return [self.exists(value) for value in values]

    def insert_many(self, values):
        return [self.insert(value) for value in values]

    def exists_and_insert(self, value):
        return self.default_stat

    def exists_and_insert_many(self, values):
        return [self.exists_and_insert(value) for value in values]

    def close(self):
        pass


class AsyncDefaultFilter(DefaultFilter):
    async def exists(self, *args, **kwargs):
        return super(AsyncDefaultFilter, self).exists(*args, **kwargs)

    async def insert(self, *args, **kwargs):
        return super(AsyncDefaultFilter, self).insert(*args, **kwargs)

    async def exists_many(self, *args, **kwargs):
        return super(AsyncDefaultFilter, self).exists_many(*args, **kwargs)

    async def insert_many(self, *args, **kwargs):
        return super(AsyncDefaultFilter, self).insert_many(*args, **kwargs)

    async def exists_and_insert(self, *args, **kwargs):
        return super(AsyncDefaultFilter, self).exists_and_insert(
            *args, **kwargs)

    async def exists_and_insert_many(self, *args, **kwargs):
        return super(AsyncDefaultFilter, self).exists_and_insert_many(
            *args, **kwargs)


class FilterCounter(object):
    def __init__(self, stats=None):
        if not stats:
            stats = []
        self.stats = [bool(stat) for stat in stats]

    def count(self, stat=True):
        result = [_stat for _stat in self.stats if _stat == bool(stat)]
        return len(result)

    def insert_stat(self, stat):
        self.stats.append(bool(stat))

    def insert_stats(self, stats):
        for stat in stats:
            self.insert_stat(stat)

    def any(self, stat=True):
        return self.count(stat) > 0

    def all(self, stat=True):
        return self.count(stat) == len(self.stats)

    def reach(self, count, stat=True):
        return self.count(stat) >= count

    def appeared(self, index, stat=True):
        if len(self.stats) >= index and self.stats[-1] == stat:
            return True
        return False


class PageFilterCounter(FilterCounter):
    def __init__(self, stats=None, max_page=None, current_page=None,
                 duplicate=False):
        super(PageFilterCounter, self).__init__(stats)
        self.max_page = max_page
        if not duplicate:
            self.max_page = None
        self.current_page = 1 if current_page == 0 else current_page

    def has_next(self):
        if not self.max_page or not self.current_page:
            return True
        if self.current_page < self.max_page:
            return True
        return False



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2024/3/13


import time

from dupfilter.filters.sql import SQLFilter


class SQLiteFilter(SQLFilter):
    def _insert_sql(self, values):
        values = [self._value_hash_and_compress(value) for value in values]
        if self.record_time:
            sql = f"""INSERT INTO {self.table}
                             (id, insert_time) VALUES (?, ?)"""
            values = [(value, int(time.time())) for value in values]
        else:
            sql = f"INSERT INTO {self.table} (id) VALUES (?)"
            values = [(value, ) for value in values]
        return sql, values



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2024/3/12


from dupfilter.filters import Filter, decorate_warning
from dupfilter.filters.sql import SQLFilter


class MySQLFilter(SQLFilter):
    pass


class AsyncMySQLFilter(SQLFilter, Filter):
    def __init__(self, pool, table,
                 record_time=False, *args, **kwargs):
        self.pool = pool
        self.table = table
        self.record_time = record_time
        Filter.__init__(self, *args, **kwargs)

    async def create_table(self):
        sql = self._create_table_sql()
        async with self.pool.acquire() as conn:
            async with conn.cursor() as cur:
                try:
                    await cur.execute(sql)
                    await conn.commit()
                    self.logger.info("创建去重表%s成功" % self.table)
                except Exception as e:
                    self.logger.info("创建去重表%s已经创建或创建失败" % self.table)
                    return

    @decorate_warning
    async def exists(self, value):
        result = await self.exists_many([value])
        return result[0]

    @decorate_warning
    async def exists_many(self, values):
        sql, new_values = self._exists_sql(values)
        async with self.pool.acquire() as conn:
            async with conn.cursor() as cur:
                await cur.execute(sql)
                result = await cur.fetchall()
                result = [res[0] for res in result]
                stats = [value in result for value in new_values]
                self._log_exists(values, new_values, stats)
                return stats

    @decorate_warning
    async def insert(self, value):
        result = await self.insert_many([value])
        return result[0]

    @decorate_warning
    async def insert_many(self, values):
        sql, values = self._insert_sql(values)
        async with self.pool.acquire() as conn:
            async with conn.cursor() as cur:
                await cur.executemany(sql, values)
                await conn.commit()
                return [True for _ in values]

    @decorate_warning
    async def exists_and_insert(self, value):
        result = await self.exists_and_insert_many([value])
        return result[0]

    @decorate_warning
    async def exists_and_insert_many(self, values):
        stats = await self.exists_many(values)
        values = [value for stat, value in zip(stats, values) if not stat]
        await self.insert_many(values)
        return stats


if __name__ == '__main__':
    import aiomysql
    import asyncio


    async def test():
        configs = {
            'minsize': 1,
            'maxsize': 10,
            'host': '192.168.1.10',
            'user': 'root',
            'password': '123456',
            'port': 3306,
            'db': 'test'
        }
        pool = await aiomysql.create_pool(**configs)
        f = AsyncMySQLFilter(pool=pool, table="dup", record_time=True)
        await f.create_table()
        res = await f.exists_and_insert_many(['1', '2', '3'])
        print(res)


    loop = asyncio.get_event_loop()
    loop.run_until_complete(test())



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/9/6
from dupfilter.filters import decorate_warning
from dupfilter.filters.redis import RedisFilter

EXISTS_SCRIPT = """
local keys = KEYS
local expire = ARGV[1]
local result = {}
for _, key in pairs(keys) do
    local exist = redis.call('SETNX', key, '')
    if exist==1 then
        table.insert(result, 0)
        redis.call('EXPIRE', key, expire)
    else
        table.insert(result, 1)
    end
end
return result

"""
INSERT_SCRIPT = """
local keys = KEYS
local expire = ARGV[1]
local result = {}
for _, key in pairs(keys) do
    local exist = redis.call('EXISTS', key)
    table.insert(result, exist)
    if exist==1 then
        redis.call('EXPIRE', key, expire)
    end
end
return result
"""


class RedisStringFilter(RedisFilter):
    def __init__(self, server, *args, **kwargs):
        super(RedisStringFilter, self).__init__(server, *args, **kwargs)
        self.exists_script = self.server.register_script(EXISTS_SCRIPT)
        self.insert_script = self.server.register_script(INSERT_SCRIPT)

    @decorate_warning
    def exists(self, value, expire=7200):
        return self.exists_many([value], expire)[0]

    @decorate_warning
    def exists_many(self, values, expire=7200):
        keys = [self._value_hash_and_compress(value) for value in values]
        stats = self.exists_script(keys=keys, args=[expire])
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, keys, stats)
        return stats

    @decorate_warning
    def insert(self, value, expire=2592000):
        return self.insert_many([value], expire)[0]

    @decorate_warning
    def insert_many(self, values, expire=2592000):
        keys = [self._value_hash_and_compress(value) for value in values]
        stats = self.insert_script(keys=keys, args=[expire])
        return [bool(stat) for stat in stats]

    exists_and_lock = exists
    exists_and_lock_many = exists_many
    confirm = insert
    confirm_many = insert_many


class AsyncRedisStringFilter(RedisStringFilter):
    @decorate_warning
    async def insert(self, value, expire=2592000):
        stats = await self.insert_many([value], expire)
        return stats[0]

    @decorate_warning
    async def insert_many(self, values, expire=2592000):
        keys = [self._value_hash_and_compress(value) for value in values]
        stats = await self.insert_script(keys=keys, args=[expire])
        return [bool(stat) for stat in stats]

    @decorate_warning
    async def exists(self, value, expire=7200):
        stats = await self.exists_many([value], expire)
        return stats[0]

    @decorate_warning
    async def exists_many(self, values, expire=7200):
        keys = [self._value_hash_and_compress(value) for value in values]
        stats = await self.exists_script(keys=keys, args=[expire])
        stats = [bool(stat) for stat in stats]
        self._log_exists(values, keys, stats)
        return stats



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/9/7


import random

from dupfilter.filters import Filter, decorate_warning
from dupfilter.filters import Reset, decorate_reset


class MemoryReset(Reset):
    @property
    def current_count(self):
        return len(self.flt.dups)

    def reset(self):
        rm_items = set(random.sample(self.flt.dups, self.reset_count))
        self.flt.dups -= rm_items


class MemoryFilter(Filter):
    def __init__(self, *args, **kwargs):
        self.dups = set()
        super(MemoryFilter, self).__init__(*args, **kwargs)

    @decorate_warning
    def exists(self, value):
        new_value = self._value_hash_and_compress(value)
        stat = new_value in self.dups
        self._log_exists([value], [new_value], [stat])
        return stat

    @decorate_warning
    def exists_many(self, values):
        return [self.exists(value) for value in values]

    @decorate_warning
    @decorate_reset
    def insert(self, value):
        new_value = self._value_hash_and_compress(value)
        self.dups.add(new_value)
        return True

    @decorate_warning
    def insert_many(self, values):
        return [self.insert(value) for value in values]

    @decorate_warning
    @decorate_reset
    def exists_and_insert(self, value):
        new_value = self._value_hash_and_compress(value)
        stat = new_value in self.dups
        self._log_exists([value], [new_value], [stat])
        if not stat:
            self.dups.add(new_value)
        return stat

    @decorate_warning
    def exists_and_insert_many(self, values):
        return [self.exists_and_insert(value) for value in values]



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2023/9/7


import os
import random

from dupfilter.filters import decorate_warning, decorate_reset
from dupfilter.filters.memory import MemoryFilter
from dupfilter.filters.memory import MemoryReset


class FileReset(MemoryReset):

    def reset(self):
        super(FileReset, self).reset()
        self.flt.file.truncate(0)
        for dup in self.flt.dups:
            self.flt.file.write(dup + '\n')


class FileFilter(MemoryFilter):
    def __init__(self,
                 path='./',
                 *args, **kwargs):
        self.path = path
        self.file = open(os.path.join(self.path, 'dup'), 'a+')
        self.file.seek(0)
        super(FileFilter, self).__init__(*args, **kwargs)
        self.dups.update(x.rstrip() for x in self.file)

    @decorate_reset
    def _insert(self, value):
        self.dups.add(value)
        self.file.write(value + '\n')

    @decorate_warning
    def insert(self, value):
        new_value = self._value_hash_and_compress(value)
        if new_value not in self.dups:
            self._insert(new_value)
        return True

    @decorate_warning
    def exists_and_insert(self, value):
        new_value = self._value_hash_and_compress(value)
        stat = new_value in self.dups
        self._log_exists([value], [new_value], [stat])
        if not stat:
            self._insert(new_value)
        return stat



# -*- coding:utf-8 -*-
# author: kusen
# email: 1194542196@qq.com
# date: 2024/3/12


import time
from dupfilter.filters.sql import SQLFilter
from dupfilter.filters.sql.mysql import AsyncMySQLFilter


class OracleSQLFilter(SQLFilter):

    def _exists_table(self):
        sql = f"SELECT TABLE_NAME FROM USER_TABLES WHERE TABLE_NAME = '{self.table.upper()}'"
        try:
            self.cursor.execute(sql)
            res = self.cursor.fetchone()
            return bool(res)
        except:
            return False

    def _insert_sql(self, values):
        values = [self._value_hash_and_compress(value) for value in values]
        if self.record_time:
            sql = f"""INSERT /*+ IGNORE_ROW_ON_DUPKEY_INDEX(
            {self.table}(id)) */ INTO {self.table}
                     (id, insert_time) VALUES (:1,:2)"""
            values = [(value, int(time.time())) for value in values]
        else:
            sql = f"""INSERT /*+ IGNORE_ROW_ON_DUPKEY_INDEX(
            {self.table} INTO {self.table} (id) VALUES (:1)"""
        return sql, values


class AsyncOracleSQLFilter(OracleSQLFilter, AsyncMySQLFilter):
    def __init__(self, *args, **kwargs):
        AsyncMySQLFilter.__init__(self, *args, **kwargs)


